#include "vl.h"
#define DEBUG

///////////////////////////////////////////////////////////////////////////////
// k_open() - Behaves exactly like the original sys_open() except assumes the
//            filename is in kernel space, not user space
///////////////////////////////////////////////////////////////////////////////
int k_open(char *filename, int flags, int mode)
{
	int fd;

	lock_kernel();
	fd = get_unused_fd();
	if (fd >= 0) {
		struct file *f = filp_open(filename, flags, mode);
		if(IS_ERR(f)) {

			// The following statement is equivalent to put_unused_fd(fd);
			//
		        FD_CLR(fd, &current->files->open_fds);
			fd = PTR_ERR(f);
		}
		else fd_install(fd, f);
	}
	unlock_kernel();
	return fd;
}


///////////////////////////////////////////////////////////////////////////////
// vl_open() - Replacement for sys_open()
//    Calls sys_open() to perform the open.  For normal files, checks if the
//    parent directories versioning policies are known.  If not, it attempts
//    to read the policies from the .version file in the parent directory.
//    If unsuccessful, versioning is not performed.  No versioning is
//    performed on the .version file itself.
///////////////////////////////////////////////////////////////////////////////
asmlinkage int vl_open(const char *filename, int flags, int mode)
{
	int fdesc, status;
	struct file *fp;
	struct dentry *fdent;
	struct dpolicy *dpol;
	char *kfname, *nvfile = NULL;
	//char *path, *cptr;
	//struct verinfo vinf;
	//int idx, begin, sub, mul, rel, ver;

	kfname = getname(filename);

	// Relative version numbering support: Needs sys_readdir, etc support
	//
	//if(match(kfname, "\\SS*,-\\dd*|?*/\\SS*,-\\dd*")) {
	//	idx = 0;
	//	cptr = kfname;
	//	begin = 0;
	//	while(*cptr) {
	//		if(*cptr++ == '/') begin = idx;
	//		idx++;
	//	}
	//	if(!begin) {
	//		rel = 1;
	//	}
	//	else {
	//		path = __getname();
	//		strcpy(path, kfname);
	//		path[begin + 1] = 0;
	//		begin += 1;
	//		rel = 0;
	//		
	//	}
	//	strcpy(kfname, &(filename[begin]));
	//	idx = strlen(kfname);
	//	mul = 1;
	//	sub = 0;
	//	while(kfname[--idx] != '-') {
	//		sub -= (kfname[idx] - '0') * mul;
	//		mul *= 10;
	//	}
	//	kfname[idx - 1] = 0;		
	//	if(rel) {
	//		get_version_number(&vinf, current->fs->pwd, kfname);
	//		ver = sub + vinf.hi + 1;
	//	}
	//	else {
	//		fdent = lookup_dentry(path, NULL, 1);
	//		get_version_number(&vinf, fdent, kfname);
	//		dput(fdent);
	//		ver = sub + vinf.hi + 1;
	//		putname(path);
	//	}
	//	if(!vinf.num || (ver < 0)) strcpy(kfname, filename);
	//	else {
	//		cptr = __getname();
	//		if(rel) sprintf(cptr, ".%s,%d", kfname, ver);
	//		else sprintf(cptr, "%s.%s,%d", path, kfname, ver);
	//		strcpy(kfname, cptr);
	//		putname(cptr);
	//		printk("Final filename: %s\n", kfname);
	//	}
	//}
	//strcpy(kfname, filename);

	if(flags & O_CREAT) {
		fdent = lookup_dentry(kfname, NULL, (flags & O_EXCL ? 0 : 1));
		if(IS_ERR(fdent)) {
			putname(kfname);
			return PTR_ERR(fdent);
		}
		if(!fdent->d_inode) {
			dput(fdent);
			status = k_open(kfname, flags, mode);
			putname(kfname);
			return status;
		}
		dput(fdent);
	}

	if((fdesc = k_open(kfname, flags & ~O_TRUNC, mode)) < 0) {
		putname(kfname);
		return fdesc;
	}
	fp = fcheck(fdesc);
	
	// Need not worry about versioning if the file is opened
	// read-only (except that file can be truncated even when opened
	// read-only!)
	//
	if(!(flags & O_TRUNC) && ((flags & O_ACCMODE) == O_RDONLY)) {
		putname(kfname);
		return fdesc;
	}

	// If the file being opened is not a regular file (e.g. a pipe or a
        // device, do not apply versioning
	//
	if(!S_ISREG(fp->f_dentry->d_inode->i_mode)) goto no_version;

#ifdef DEBUG
	printk(KERN_INFO "open(%s, %s%s%s) = %d\n", kfname,
	       (flags & O_TRUNC ? "TRUNC " : ""),
	       (flags & O_CREAT ? "CREAT " : ""),
	       (flags & O_EXCL ? "EXCL " : ""),
	       fdesc);
#endif

	// Do not allow versioning of the policy file!!  However, still place
	// an entry in the file policy cache so if a .version file is
	// modified, we can adjust the directory policy cache
	// **FIXME: Why Not? Allow Versioning AND adjust directory policy
	//          cache on close.  Moreover: read the .version file being
	//          opened so that if another file in that directory is opened
	//          later, we don't try to read an inconsistent .version file
	//
	if(!strcmp(fp->f_dentry->d_name.name, vfile)) {
		struct fpolicy *vfpol;

		vfpol = ins_fpolicy(fp, NULL);
		if(vfpol) {
			vfpol->ispolicyfile = 1;
			if(flags & O_TRUNC) vfpol->isdirty = 1;
		}
		goto no_version;
	}

	dpol = get_dpolicy(fp->f_dentry->d_parent);

	if(dpol->onwrite) {
		int incl = 0;

		if(dpol->isize &&
		   match(fp->f_dentry->d_name.name, dpol->incl)) {
#ifdef DEBUG
			printk("    INCLUDE regex match!\n");
#endif
			incl = 1;
		}
		if(!incl && dpol->esize &&
		   match(fp->f_dentry->d_name.name, dpol->excl)) {
#ifdef DEBUG
			printk("    EXCLUDE regex match!\n");
#endif
			goto no_version;
		}

#ifdef DEBUG
		printk("    Versioning %s\n", kfname);
#endif
		ins_fpolicy(fp, dpol);

		if(flags & O_TRUNC) {
			nvfile = __getname();
			if(!nvfile) goto no_version;
			status = create_version(fdesc, nvfile);
			if(status < 0) {
				printk("VL: create_version() fails, err=%d\n", status);
				putname(nvfile);
				nvfile = NULL;
			}
#ifdef DEBUG
			else printk("    TRUNC, created %s as new version\n",
				    nvfile);
#endif
		}
	}

 no_version:
	if(flags & O_TRUNC) {
		mode_t old_mode;

		old_mode = fp->f_mode;
		fp->f_mode |= FMODE_WRITE;
		status = old_sys_ftruncate(fdesc, 0);
		fp->f_mode = old_mode;

		if(status < 0) {
			if(nvfile) {
#ifdef DEBUG
				printk("    ftruncate() failed, err %d. "
				       "Unlinking %s\n", status, nvfile);
#endif
				status = old_sys_unlink(nvfile);
				if(status < 0)
					printk("VL: unlink() following failed "
					       "ftruncate() failed: %d\n",
					       status);
				putname(nvfile);
			}
			old_sys_close(fdesc);
			putname(kfname);
			return status;
		}
#ifdef DEBUG
		printk("    Truncated file %s\n", kfname);
#endif
	}
	putname(kfname);
	return fdesc;
}

///////////////////////////////////////////////////////////////////////////////
// vl_creat() - Replacement for sys_creat()
///////////////////////////////////////////////////////////////////////////////
asmlinkage int vl_creat(const char *pathname, int mode)
{
	return vl_open(pathname, O_CREAT | O_WRONLY | O_TRUNC, mode);
}

///////////////////////////////////////////////////////////////////////////////
// vl_close() - Replacement for sys_close()
///////////////////////////////////////////////////////////////////////////////
asmlinkage int vl_close(unsigned int fd)
{
	struct file *fp;
	struct fpolicy *fpol;

	if((fp = fcheck(fd)) == NULL) goto do_close;
	fpol = get_fpolicy(fp);
	if(!fpol) goto do_close;

#ifdef DEBUG
	printk("vl_close(%d)\n", fd);
#endif

	// If the file closed was a .version file and it is dirty, invalidate
	// the policy cache entry on that directory (forcing .version) to be
	// re-read
	//
	if(fpol->ispolicyfile && fpol->isdirty) {
#ifdef DEBUG
		printk("    Closed dirty .version file\n");
#endif
		invalidate_dpolicy(fp->f_dentry->d_parent);
	}

	// Remove the entry in the file policy cache for this file
	//
	rem_fpolicy(fp);

 do_close:

	// Close the file
	//
	return old_sys_close(fd);
}

///////////////////////////////////////////////////////////////////////////////
// vl_unlink() - Replacement for sys_unlink()
///////////////////////////////////////////////////////////////////////////////
asmlinkage int vl_unlink(const char *pathname)
{
	char *kfname, *newname;
	int incl, status;
	struct dentry *fdent, *old_wd;
	struct dpolicy *dpol;
	struct verinfo vinf;
	mm_segment_t old_fs;

	kfname = getname(pathname);

	fdent = lookup_dentry(kfname, NULL, 0);
	if(IS_ERR(fdent)) {
		putname(kfname);
		return PTR_ERR(fdent);
	}

#ifdef DEBUG
	printk("unlink(%s)\n", pathname);
#endif

	if(!fdent->d_inode) goto do_unlink;
	if(!fdent->d_inode->i_size) goto do_unlink;
	if(!S_ISREG(fdent->d_inode->i_mode)) goto do_unlink;

	if(!strcmp(fdent->d_name.name, vfile)) {
#ifdef DEBUG
		printk("    Unlinking .version file\n");
#endif
		invalidate_dpolicy(fdent->d_parent);
	}

	dpol = get_dpolicy(fdent->d_parent);
	
	if(!dpol->onunlink) goto do_unlink;
	if(match(fdent->d_name.name, ".??*,\\dd*")) goto do_unlink;
	if(dpol->isize && match(fdent->d_name.name, dpol->incl)) incl = 1;
	else incl = 0;
	if(!incl && dpol->esize && match(fdent->d_name.name, dpol->excl)) goto do_unlink;
	get_version_number(&vinf, fdent->d_parent, pathname);

	newname = __getname();
	old_wd = current->fs->pwd;
	current->fs->pwd = fdent->d_parent;
	sprintf(newname, ".%s,%d", fdent->d_name.name, vinf.hi + 1);
	current->fs->pwd = old_wd;

	old_fs = get_fs();
	set_fs(KERNEL_DS);

	cleanup(fdent, dpol->numlimit - 1, &vinf, NULL);

#ifdef DEBUG
	printk("    Renaming %s to %s\n", pathname, newname);
#endif
	status = old_sys_rename(pathname, newname);
	set_fs(old_fs);

	if(dpol->rdonly) {
		fdent->d_inode->i_mode = (fdent->d_inode->i_mode & ~S_IALLUGO) |
			(fdent->d_inode->i_mode & (S_IRUSR | S_IRGRP | S_IROTH));
	}

	putname(newname);
	dput(fdent);
	putname(kfname);
	return status;

 do_unlink:
	dput(fdent);
	putname(kfname);
	return old_sys_unlink(pathname);
}

asmlinkage int vl_rename(const char *oldname, const char *newname)
{
	int fd;
	char *nvfile;

#ifdef DEBUG
	printk("rename(%s, %s)\n", oldname, newname);
#endif

	fd = vl_open(newname, O_WRONLY, 0);
	if(fd < 0) {
#ifdef DEBUG
		printk("    Error opening %s in O_WRONLY mode\n", newname);
#endif
		return old_sys_rename(oldname, newname);
	}

	nvfile = __getname();
	create_version(fd, nvfile);
	putname(nvfile);

	vl_close(fd);
	
	return old_sys_rename(oldname, newname);
}

asmlinkage int vl_truncate(const char *path, unsigned long length)
{
	int fd, status;

#ifdef DEBUG
	printk("truncate(%s, %ld)\n", path, length);
#endif

	fd = vl_open(path, O_WRONLY, 0);
	if(fd < 0) {
#ifdef DEBUG
		printk("    Error opening %s in mode O_WRONLY\n", path);
#endif
		return fd;
	}

	status = vl_ftruncate(fd, length);
	old_sys_close(fd);
	return status;
}

asmlinkage int vl_ftruncate(unsigned int fd, unsigned long length)
{
	struct file *fp;
	char *nvfile;

#ifdef DEBUG
	printk("ftruncate(%d, %ld)\n", fd, length);
#endif

	fp = fcheck(fd);
	if(!fp) return -EBADF;

	if(length >= fp->f_dentry->d_inode->i_size) goto do_ftrunc;

	nvfile = __getname();
	create_version(fd, nvfile);
	putname(nvfile);

 do_ftrunc:
	return old_sys_ftruncate(fd, length);
}
