#include "vl.h"
#define DEBUG

// Versioning policy filename
//
char *vfile = ".version\0";

// Policy cache
//
struct dpolicy ptab[DPOL_TABLE_SIZE];

// Open file policies
//
struct fpolicy *fphead;

struct dpolicy *get_dpolicy(struct dentry *dir)
{
	struct inode *dinode;
	unsigned short idx;

	// Lookup the directory in the cache to see if
	// there are known versioning policies for that directory
	//
	dinode = dir->d_inode;
	idx = (unsigned short)(((long)dinode >> 2) & DPOL_MASK);

	// If the entry is tagged with the full inode pointer for the file,
	// then we have the policies for this directory cached.  Otherwise, we
	// need to read the .version file to find the policies
	//
	if((ptab[idx].tag_inode == dinode->i_ino) && (ptab[idx].tag_device == dinode->i_dev)) {
#ifdef DEBUG
		printk("    Policy cache HIT (%d)\n", idx);
#endif
		return &(ptab[idx]);
	}

#ifdef DEBUG
	printk("    Policy cache MISS (%d)\n", idx);
#endif
	
	// If a valid entry is being replaced, free its list of
	// overrides
	//
	if(ptab[idx].isize) {
		kfree(ptab[idx].incl);
		ptab[idx].isize = 0;
	}
	if(ptab[idx].esize) {
		kfree(ptab[idx].excl);
		ptab[idx].esize = 0;
	}
	
	// Cache the policies for this directory
	// 736
	ptab[idx].tag_inode = dinode->i_ino;
	ptab[idx].tag_device = dinode->i_dev;

	if(read_dpolicy(dir, &(ptab[idx]))) {
#ifdef DEBUG
		printk("    No versioning policies exist\n");
#endif

		ptab[idx].onunlink = 0;
		ptab[idx].onwrite  = 0;
		ptab[idx].rdonly   = 0;
		ptab[idx].numlimit = 0;
		ptab[idx].isize = 0;
		ptab[idx].esize = 0;
		ptab[idx].incl = NULL;
		ptab[idx].excl = NULL;
	}
	return &(ptab[idx]);
}

int read_dpolicy(struct dentry *dir, struct dpolicy *dpol)
{
	int vfd;
	int bytes;
	mm_segment_t old_fs;
	struct dentry *pwd;

	// Change the cwd
	//
	pwd = current->fs->pwd;
	current->fs->pwd = dir;
	
	// Open the policy file
	//
	vfd = k_open(vfile, O_RDONLY, 0);

	// Revert to the old cwd
	//
	current->fs->pwd = pwd;

	if(vfd < 0) return -1;
		
	// Allow writing to kernel memory in the following system calls
	//
	old_fs = get_fs();
	set_fs(KERNEL_DS);

	bytes = sys_read(vfd, (char *)DPOL_DATA(dpol), POLICY_DATA_SIZE);
	if(bytes != POLICY_DATA_SIZE) {
		printk("VL: Error reading policy file: %d\n", bytes);
		sys_close(vfd);
		set_fs(old_fs);
		return -1;
	}
	if(dpol->isize) {
		dpol->incl = (char *)kmalloc(dpol->isize, GFP_KERNEL);
		if(!dpol->incl) {
			printk("VL: kmalloc() error\n");
			sys_close(vfd);
			set_fs(old_fs);
			return -1;
		}
		bytes = sys_read(vfd, dpol->incl, dpol->isize);
		if(bytes != dpol->isize) {
			kfree(dpol->incl);
			printk("VL: Error reading policy file: %d\n", bytes);
			sys_close(vfd);
			set_fs(old_fs);
			return -1;
		}
	}
	if(dpol->esize) {
		dpol->excl = (char *)kmalloc(dpol->esize, GFP_KERNEL);
		if(!dpol->excl) {
			printk("VL: kmalloc() error\n");
			sys_close(vfd);
			set_fs(old_fs);
			return -1;
		}
		bytes = sys_read(vfd, dpol->excl, dpol->esize);
		if(bytes != dpol->esize) {
			kfree(dpol->excl);
			printk("VL: Error reading policy file: %d\n", bytes);
			sys_close(vfd);
			set_fs(old_fs);
			return -1;
		}
	}
	sys_close(vfd);
	
	// Disallow writing to kernel memory
	//
	set_fs(old_fs);

	return 0;
}

void invalidate_dpolicy(struct dentry *dir)
{
	struct inode *dinode;
	unsigned short idx;

	dinode = dir->d_inode;
	idx = (unsigned short)(((long)dinode >> 2) & DPOL_MASK);

	if((ptab[idx].tag_inode == dinode->i_ino) && (ptab[idx].tag_device = dinode->i_dev)) {
#ifdef DEBUG
		printk("    Invalidating policy cache entry\n");
#endif
		if(ptab[idx].isize) kfree(ptab[idx].incl);
		if(ptab[idx].esize) kfree(ptab[idx].excl);
		ptab[idx].tag_inode = 0;
		ptab[idx].tag_device = 0;
		ptab[idx].isize = 0;
		ptab[idx].esize = 0;
	}
}

struct fpolicy *ins_fpolicy(struct file *file, struct dpolicy *dpol)
{
	struct fpolicy *curr;

	// Search the list of open files to find out if the file is already
	// open
	//
	if(fphead) {
		curr = fphead;

		do {
			if(curr->tag == file) {
				// FIXME: Verify dpolicies are same as
				//        currently in list or print error
				//
				return curr;
			}
		} while((curr = curr->next));
	}

	// Allocate new fpolicy structure
	//
	curr = (struct fpolicy *)kmalloc(sizeof(struct fpolicy), GFP_KERNEL);
	if(!curr) return NULL;
	curr->tag = file;

	// Copy the directory policies to that of the file
	//
	curr->isdirty = 0;
	curr->ispolicyfile = 0;
	if(dpol) {
		curr->numlimit = dpol->numlimit;
		curr->rdonly = dpol->rdonly;
	}
	else {
		curr->numlimit = 0;
		curr->rdonly = 0;
	}
	
	// Insert the new structure in the list
	//
	curr->next = fphead;
	fphead = curr;

	return curr;
}

int rem_fpolicy(struct file *file)
{
	struct fpolicy *curr, *prev;

	if(!fphead) return -1;
	prev = fphead;
	if(fphead->tag == file) {
		if(file->f_count > 1) return 0;
		fphead = fphead->next;
		kfree(prev);
		return 0;
	}

	while((curr = prev->next)) {
		if(curr->tag == file) {
			if(file->f_count > 1) return 0;
			prev->next = curr->next;
			kfree(curr);
			return 0;
		}
		prev = curr;
	}
	return -1;
}

struct fpolicy *get_fpolicy(struct file *file)
{
	struct fpolicy *curr;

	if(!fphead) return NULL;
	curr = fphead;

	do {
		if(curr->tag == file) return curr;
	} while((curr = curr->next));

	return NULL;
}



