#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

struct dpolicy {
	unsigned long tag_inode;
	unsigned short tag_device;
	unsigned char numlimit, isize, esize;
	unsigned char onwrite  : 1;
	unsigned char onunlink : 1; 
	unsigned char rdonly   : 1;
	char *incl, *excl;
};


#define LIMIT  1
#define INCL   2
#define EXCL   4
#define UNLINK 8
#define WRITE  16
#define RDONLY 32

#define POLICY_DATA_SIZE (sizeof(struct dpolicy) - sizeof(unsigned long) - \
			  sizeof(unsigned short) - (2 * sizeof(char *)))

int optind, opterr, optopt;
char *optarg;

void main(int argc, char **argv)
{
	int fd, stat, dirty, num;
	struct dpolicy dpol;
	char incl[80];
	char excl[80];

	if((fd = open(".version", O_RDONLY)) != -1) {
		if(read(fd, &(dpol.numlimit), POLICY_DATA_SIZE) != POLICY_DATA_SIZE)
			goto filerr;
		if(dpol.isize) if(read(fd, incl, dpol.isize) != dpol.isize) goto filerr;
		if(dpol.esize) if(read(fd, excl, dpol.esize) != dpol.esize) goto filerr;
		close(fd);

	}
	else {
		dpol.numlimit = 0;
		dpol.onunlink = 0;
		dpol.onwrite = 0;
		dpol.rdonly = 0;
		dpol.isize = 0;
		dpol.esize = 0;
		incl[0] = 0;
		excl[0] = 0;
	}

	dirty = 0;

	do {
		opterr = 0;
		stat = getopt(argc, argv, "n:i:e:wWdDrRh");
		switch(stat) {
		case 'n':
			num = atoi(optarg);
			if(num < 0) {
				printf("Error: must specify a non-negative version limit\n");
				exit(1);
			}
			if(num != dpol.numlimit) dirty |= LIMIT;
			dpol.numlimit = atoi(optarg);
			if(!num) {
				if((dirty & WRITE) && dpol.onwrite) {
					printf("Error: may not specify -n0 and -W\n");
					exit(1);
				}
				if((dirty & UNLINK) && dpol.onunlink) {
					printf("Error: may not specify -n0 and -D\n");
					exit(1);
				}
				dpol.onwrite = 0;
				dpol.onunlink = 0;
				dirty |= (WRITE | UNLINK);
			}
			break;
		case 'i':
			if(strcmp(incl, optarg)) dirty |= INCL;
			strncpy(incl, optarg, 79);
			dpol.isize = strlen(incl) + 1;
			break;
		case 'e':
			if(strcmp(excl, optarg)) dirty |= EXCL;
			strncpy(excl, optarg, 79);
			dpol.esize = strlen(excl) + 1;
			break;
		case 'w':
			if((dirty & WRITE) && dpol.onwrite) {
				printf("Error: may not specify both -w and -W\n");
				exit(1);
			}
			if(dpol.onwrite) {
				dpol.onwrite = 0;
				dirty |= WRITE;
			}
			break;
		case 'W':
			if((dirty & WRITE) && !dpol.onwrite) {
				printf("Error: may not specify both -W and -w or -n0\n");
				exit(1);
			}
			if(!dpol.onwrite) {
				dpol.onwrite = 1;
				dirty |= WRITE;
			}
			break;
		case 'd':
			if((dirty & UNLINK) && dpol.onunlink) {
				printf("Error: may not specify both -d and -D\n");
				exit(1);
			}
			if(dpol.onunlink) {
				dpol.onunlink = 0;
				dirty |= UNLINK;
			}
			break;
		case 'D':
			if((dirty & UNLINK) && !dpol.onunlink) {
				printf("Error: may not specify both -D and -d or -n0\n");
				exit(1);
			}
			if(!dpol.onunlink) {
				dpol.onunlink = 1;
				dirty |= UNLINK;
			}
			break;
		case 'r':
			if((dirty & RDONLY) && dpol.rdonly) {
				printf("Error: may not specify both -r and -R\n");
				exit(1);
			}
			if(dpol.rdonly) {
				dpol.rdonly = 0;
				dirty |= RDONLY;
			}
			break;
		case 'R':
			if((dirty & RDONLY) && !dpol.rdonly) {
				printf("Error: may not specify both -R and -r\n");
				exit(1);
			}
			if(!dpol.rdonly) {
				dpol.rdonly = 1;
				dirty |= RDONLY;
			}
			break;
		case '?':
		case 'h':
			printf("Usage: setpol [options]\n");
			printf("Without any options, setpol displays the versioning policies\n");
			printf("in the current working directory.  The following command line\n");
			printf("options may be used to change the versioning policies.  After\n");
			printf("changing the policies, setpol displays the new policies.\n");
			printf("\t-W          : Creates a new version when a file is written\n");
			printf("\t-w          : Does not create a new version on file write\n");
			printf("\t              Only one of -W and -w is allowed.\n");
			printf("\t-D          : Creates a new version when a file is deleted\n");
			printf("\t-d          : Does not create a new version on file deletion\n");
			printf("\t              Only one of -D and -d is allowed.\n");
			printf("\t-R          : Access mode of new version files is cleared of\n");
			printf("\t              all write/execute/setuid/setgid bits\n");
			printf("\t-r          : Access mode of version files is set to access\n");
			printf("\t              mode of original file\n");
			printf("\t              Only one of -R and -r is allowed.\n");
			printf("\t-n [number] : Limits the number of old versions kept per file.\n");
			printf("\t              Setting zero implies -w and -d.\n");
			printf("\t-e [regex]  : Excludes files matching regex from being versioned\n");
			printf("\t-i [regex]  : Files matching regex are included even if the file\n");
			printf("\t              would be excluded by -e\n");
			printf("Regular expressions are of the format [expr1|expr2|expr3|...|exprn].  Each\n");
			printf("subexpression exprn is compared character wise against the filename to determine\n");
			printf("if the pattern matches.  The following special characters may also be used:\n");
			printf("\t?  : Matches any character\n");
			printf("\t\\d : Matches any digit (0-9)\n");
			printf("\t\\\\ : Matches a backslash\n");
			printf("\t?* : Matches as many characters as possible\n");
			printf("\td* : Matches as many digits as possible\n");
			printf("\tx* : Any other character x prior to an * behaves as ?*\n");
			printf("There is a limit of three * per subexpression or that subexpression may\n");
			printf("not match correctly.\n");
			exit(0);
		}

	} while(stat != EOF);

	if((dpol.onwrite || dpol.onunlink) && !dpol.numlimit) {
		printf("Error: must specify a limit of at least 1 when -W or -D is specified\n");
		exit(1);
	}

	printf("Current policies:\n");
	if(dpol.onwrite || dpol.onunlink) {
		if(dpol.numlimit == 1) printf("\tVersioning is ON, limit of 1 old version\n");
		else printf("\tVersioning is ON, limit of %d old versions\n", dpol.numlimit);

		if(dpol.onwrite) printf("\tFiles will be versioned when written\n");
		else printf("\tFiles will NOT be versioned when written\n");

		if(dpol.onunlink) printf("\tFiles will be versioned when deleted\n");
		else printf("\tFiles will NOT be versioned when deleted\n");

		if(dpol.rdonly) printf("\tOld versions will be made read only\n");
		else printf("\tOld versions will have permissions of original file\n");

		if(dpol.isize) printf("\tFiles matching '%s' are always included\n", incl);
		if(dpol.esize) printf("\tFiles matching '%s' are excluded\n", excl);
	}
	else printf("\tVersioning is OFF\n");

	if(!dirty) exit(0);

	if((fd = open(".version", O_WRONLY | O_TRUNC | O_CREAT, S_IRUSR | S_IWUSR)) == -1)
		goto openerr;

	if(write(fd, &(dpol.numlimit), POLICY_DATA_SIZE) != POLICY_DATA_SIZE)
		goto filerr;

	if(dpol.isize) {
		if(write(fd, incl, dpol.isize) != dpol.isize)
			goto filerr;
	}

	if(dpol.esize) {
		if(write(fd, excl, dpol.esize) != dpol.esize)
			goto filerr;
	}

	close(fd);
	exit(0);

 filerr:
	perror("File operation");
	close(fd);
	exit(1);

 openerr:
	perror("open()");
	exit(1);
}





