#include "vl.h"
#include <limits.h>

#define DEBUG

struct vl_dirscan_t {
	const char *name;
	struct verinfo *info;
};


// This is the callback that is called by the filesystem
// for each directory entry found during readdir
//
static int vl_filldir(void * __buf, const char *direntry, int namlen, off_t offset, ino_t ino) {

	struct vl_dirscan_t *dscan = (struct vl_dirscan_t *)__buf;
	const char *myname = dscan->name;
	int vernum=0;

	if (*direntry != '.') return 0;
	direntry++; namlen--;

	while (*myname && namlen) {
		if (*(direntry++) != *(myname++)) return 0;
		namlen--;
	}
	if (namlen==0) return 0;

	if (*(direntry) != ',') return 0;
	namlen--; direntry++;

	while (namlen--) {
		if ((*direntry < '0') || (*direntry > '9')) return 0;
		vernum = vernum*10 + (int)*direntry - '0';
		direntry++;
	}

	if (vernum > dscan->info->hi) dscan->info->hi = vernum;
	if (vernum < dscan->info->lo) dscan->info->lo = vernum;
	dscan->info->num++;
	return 0;
}

#undef memset
void* memset(void *_p, int value, __kernel_size_t sz) {
	int i;
	char *p=(char *)_p;
	for (i=0; i<sz; i++, p++) *p=value;
	return _p;
}

int get_version_number(struct verinfo *vinf, struct dentry *dentry, const char *name) {

	struct file file;
	struct inode *inode;
	struct vl_dirscan_t dscan;
	int error;

	error = -EBADF;

	if (!dentry) goto out_putf;
	inode = dentry->d_inode;
	if (!inode) goto out_putf;


	// Open directory as file
	//
	memset((void*)&file, 0, sizeof(file));
	file.f_count = 1;
	//file.f_version = ++event;
        file.f_dentry = dentry;
	file.f_pos = 0;
	file.f_reada = 0;
	file.f_op = NULL;
	if (inode->i_op)
		file.f_op = inode->i_op->default_file_ops;
	if (file.f_op && file.f_op->open) {
		error = file.f_op->open(inode,&file);
		if (error) goto out_putf;
	}
	file.f_flags = O_RDONLY;


	// Prepare for reading directory
	//
	error = -ENOTDIR;
	if (!file.f_op || !file.f_op->readdir) goto out_putf;
	dscan.name = name;
	vinf->hi = vinf->num = 0;
	vinf->lo = INT_MAX;
	dscan.info = vinf;

	// Get the inode's semaphore to prevent changes
	// to the directory while we read it.
	//
	down(&inode->i_sem);
	error = file.f_op->readdir(&file, &dscan, vl_filldir);
	up(&inode->i_sem);

	if (error < 0) goto out_putf;
	if (vinf->num == 0) vinf->hi = vinf->lo = -1;

#ifdef DEBUG
	printk("    Versions for %s: max = %d, min = %d, num = %d\n", name, vinf->hi, vinf->lo, vinf->num);
#endif
	return 0;

	
 out_putf:
	printk("VL: get_version_number: Error %d\n", error);
	return error;
}


// Makes a version of the file represented by fd.
// filename must point to an unused region of memory
// big enough to contain the name of the new version
// 
int create_version(unsigned int fd, char *filename) {

	char *buf, *wr_ptr;
        int  err=0, copyfd, retval=0, no_overwriting;
        mm_segment_t old_fs;
	loff_t fpos;
	mode_t oldmode;
	struct file *fp;
	struct dentry *pwd;
	struct fpolicy *pol;
	struct verinfo vinf;
	umode_t modemask;

	fp = fcheck(fd);
	if (fp==NULL) {
#ifdef DEBUG
		printk("    create_version: Bad fd (%d)\n", fd);
#endif
		return -EBADF;
	}

	//FIXME: Does it matter if <fd> is not a regular file?
	if(!S_ISREG(fp->f_dentry->d_inode->i_mode)) return 0;

	// Check Policies
	// FIXME: Should checking policies be done here?
	// It would be more apropriate at a higher level
	//
	pol = get_fpolicy(fp);
	if (pol==NULL) return 0;
	if (pol->isdirty) return 0;
	pol->isdirty = 1;
	if (pol->ispolicyfile) return 0;

	// Construct name of version
	//
	err = get_version_number(&vinf, fp->f_dentry->d_parent,
				 fp->f_dentry->d_name.name);
	if (err != 0) {
		printk("VL: get_version_number() returned error %d\n", err);
		return err;
	}
	(vinf.num==0)?vinf.hi=0:vinf.hi++;
	sprintf(filename, ".%s,%d", fp->f_dentry->d_name.name, vinf.hi);


	// Maybe the file already exists....  Should I overwrite it?
	// FIXME: This should be a policy
	// The file can only exist if multiple processors have access
	// the filesystem concurrently
	//
	no_overwriting = 0;

	// Open the new version file in the correct directory
	//
        modemask = (pol->rdonly ? (S_IRUSR | S_IRGRP | S_IROTH) : S_IALLUGO);

	pwd = current->fs->pwd;
	current->fs->pwd = fp->f_dentry->d_parent;
	copyfd = k_open(filename, O_WRONLY|O_CREAT|(O_EXCL & no_overwriting),
			      fp->f_dentry->d_inode->i_mode & modemask);
	current->fs->pwd = pwd;

	if (copyfd < 0) {
		if (copyfd == -EEXIST)
			printk("VL: create_version(): File exists: %s\n", filename);
		else 
			printk("VL: create_version(): Can't open %s; Error %d\n",
			       filename, copyfd);
		return copyfd;
	}

	// Allow access to kernel memory
	//
       	old_fs = get_fs();
       	set_fs(KERNEL_DS);


	// Seek to the beginning of source file
	// and ensure we have read permissions
	oldmode = fp->f_mode;
	fp->f_mode |= FMODE_READ;

	// Allocate Buffer
	buf = __getname();

	// Copy
	fpos = fp->f_pos;
	err = sys_lseek(fd, 0, 0);
	if (err < 0)
		printk("VL: Can't seek to beginning of file \"%s\"; Error %d\n",
		       fp->f_dentry->d_name.name, err);
	else do {
		retval = err = sys_read(fd, (void*)buf, PAGE_SIZE);
		if (err < 0) {
			printk("VL: Can't read from file \"%s\"; "
			       "Error %d\n", fp->f_dentry->d_name.name, err);
			goto out_of_copy_loop;
		}
		if (err==0) break;
		
		wr_ptr = buf;
		while (1) {
			err = old_sys_write(copyfd, (void*)(wr_ptr), retval);
			if (err < 0) {
				printk("VL: Can't write to file \"%s\"; Error %d\n",
				       filename, err);
				goto out_of_copy_loop;
			}
			if ((retval-=err) == 0) break;
			wr_ptr += err;
		}
	} while (1);

out_of_copy_loop:
	retval = err;

	// If copy succeded, try to delete old versions
	if (retval >= 0) {
		vinf.num++;
		err = cleanup(fp->f_dentry, pol->numlimit, &vinf, buf);
		if (err < 0) {
			printk("VL: cleanup(\"%s\") failed; Error: %d\n",
			       fp->f_dentry->d_name.name, err);
			if (retval >= 0) retval=err;
		}
	}

	err = old_sys_close(copyfd);
	if (err < 0) {
		printk("VL: Couldn't close \"%s\"; Error %d\n", filename, err);
		if (retval >= 0) retval=err;
	}

	err = sys_lseek(fd, fpos, 0);
	if (err < 0) {
		printk("VL: Couldn't seek to original file location (%s); "
			    "Error %d\n", fp->f_dentry->d_name.name, err);
		if (retval >= 0) retval=err;
	}

	/* Create a new link corresponding to the latest version? */
	/* Update flags? */
	
	set_fs(old_fs);
	fp->f_mode = oldmode;
	putname(buf);
	return (retval<0)?retval:vinf.hi;
}





